
/********************************************************

		xref.c
   
   Functions to glean various information about fibers and the
   objects they are assigned to.


********************************************************/

#include <stdio.h>
#include <strings.h>

#include "miscMath.h"
#include "motor.h"
#include "kaboom.h"
#include "plate.h"
#include "translate.h"
#include "flex.h"
#include "buttons.h"
#include "fiborg.h"
#include "nesslib.h"
#include "findIt.h"
#include "concen.h"
#include "coordConvert.h"
#include "xref.h"

extern char *getenv();


/********** local functions **********/

static int crossReference( xItem *theList, aStar *stars[], fibDB fibers[]  );

static void xItemClear( xItem *theList);
static int fiberToObject( int fiber );
static int fiberOnObject( xItem *theList, aStar *stars[], fibDB fibers[], double tolerance );

/********** end local functions **********/


void xRefClear( fibDB fibers[] )

{
   int i;

   for( i=0; i<=MAXFIBER; i++ ){
      fibers[i].assignedObject = (void*)NULL;
   }
}


/** Function to generate a cross reference from a fiber id **/

int xRefFiber(  int theFiber, xItem *theList, aStar *stars[], fibDB fibers[]  )

{
   xItemClear( theList);

   if( ( theFiber >= 0 ) && ( theFiber <= MAXFIBER ) ){
      theList->fiber = theFiber;
   }

   return( crossReference( theList, stars, fibers  ) );
}



/** Function to generate a cross reference from an object  id **/

int xRefObject( int theObject, xItem *theList, aStar *stars[], fibDB fibers[]  )

{
   xItemClear( theList);

   if( ( theObject >= 0 ) && ( theObject <= MAXSTARS ) ){
      theList->object = theObject;
   }

   return( crossReference( theList, stars, fibers  ));
}



/** Function to generate a cross reference from an slit id **/

int xRefSlit( int theSlit, xItem *theList, aStar *stars[], fibDB fibers[]  )

{
   xItemClear( theList);

   if( theSlit >= 0 ){
      theList->slit = theSlit;
   }

   return( crossReference( theList, stars, fibers ));
}



/** Function to generate a cross reference from a probe id **/

int xRefProbe( char theProbe, xItem *theList, aStar *stars[], fibDB fibers[]  )

{
   int probeInt;
   int i;

   xItemClear( theList);

   if( theProbe >= 'A' ){

      probeInt = (int)(theProbe) - (int)('A');
   }
   else{
      return( FALSE );
   }

   for( i=0; i<=MAXFIBER; i++ ){

      /* ignore fibers that aren't alpha labeled */

      if( !cableIsAlphaLabeled( fibers[i].cable ) )
	 continue;

      /* now test for a match */

      if( fibers[i].slitPosition == probeInt ){

	 /* now we really know what fiber we want, as an id number */
	 theList->fiber = i;
	 break;
      }
   }

   return( crossReference( theList, stars, fibers ));
}


static void xItemClear( xItem *theList )

{
   theList->fiber  = VOIDENTRY;
   theList->object = VOIDENTRY;
   theList->slit   = VOIDENTRY;
   theList->cable  = VOIDENTRY;
}



/* 
	    crossReference

   translate between fiber, object, and slit positions.
   This function *depends* on the fact that only one of 
   the three fields in the supplied structure will not 
   be a VOIDENTRY.   Call it with more than one set,
   and all bets are off!  If none is set, you're ok...
*/

static int crossReference( xItem *theList, aStar *stars[], fibDB fibers[]  )

{
   int i;

   int found;

   /* FIRST CASE, FIBER KNOWN --easiest */

   if( ( theList->fiber != VOIDENTRY ) && ( theList->fiber <= MAXFIBER ) ){

      theList->slit   = fibers[theList->fiber].slitPosition;

      theList->cable  = fibers[theList->fiber].cable;

      theList->object =  fiberToObject( theList->fiber );

      return( fiberOnObject( theList, stars, fibers, OBJ_DEAD_ZONE ) );

   }/* END FIRST CASE */



   /* SECOND CASE, SLIT KNOWN */

   if( theList->slit != VOIDENTRY ){

      found = 0;

      for( i=0; i<=MAXFIBER; i++ ){

	 /* ignore fibers that aren't part of the current cable */

	 if( !fiberInCable( i, CurrentCable ) )
	    continue;

	 /* now test for a match */

	 if( fibers[i].slitPosition == theList->slit ){

	    found = TRUE;

	    theList->fiber = i;
            theList->object =  fiberToObject( i );
      	    theList->cable = fibers[i].cable;

	    break;
	 }
      }

      if( !found ){
	 theList->slit = VOIDENTRY;
      }

      return( fiberOnObject( theList, stars, fibers, OBJ_DEAD_ZONE ) );

   }/* END SECOND CASE */


   /* THIRD CASE, OBJECT KNOWN */

   if( theList->object != VOIDENTRY ){

      for( i=0; i<=MAXSTARS; i++ ){

	 if( stars[theList->object] != NULL){ 	/* the object actually exists */

	    /* this is sneaky.  We know from the stars database
	       what fiber should be assigned, so we set the fiber
	       and then call fiberAtObject.  This saves writing
	       a second function to go the opposite direction.
	       ie: "objectAtFiber".*/


	    if( stars[theList->object]->button <= VOIDENTRY )
               theList->fiber = VOIDENTRY;
	    else
               theList->fiber = stars[theList->object]->button;

            if( fiberOnObject( theList, stars, fibers, OBJ_DEAD_ZONE ) ){

	       /* fiber was assigned, so fill in the slit position and cable  */

      	       theList->slit   = fibers[theList->fiber].slitPosition;
      	       theList->cable  = fibers[theList->fiber].cable;
	       return( TRUE );
	    }
	    else{

	       /* no luck, so void the fiber */
	       theList->fiber = VOIDENTRY;
	       return( FALSE );

	    }
	 }
      }

   }/* END THIRD CASE */

   return(FALSE);		/* not really necessary, but here for good measure */
}



static int fiberToObject( int fiber )

{
   if( FiBase[fiber].assignedObject == NULL ){
     return( VOIDENTRY );
   }
   else{
      return (((aStar*)(FiBase[fiber].assignedObject))->internalID );
   }
}


int fiberAtObject( int fiber, double tolerance )

{
   point translatedCoord;

   struct fiber_dbase *fib = &FiBase[fiber];


   if( fib->assignedObject == NULL ){

      /* we know enough to tell us the rest is pointless! */

      return ( VOIDENTRY );
   }

   /* apply translations if required */

   buttonToSky( fib, &translatedCoord );

   /* see if the two match up */

   if( distance( &translatedCoord,
                 &( ((aStar*)(fib->assignedObject))->specEncoderCoord) ) < tolerance ){

      return( ((aStar*)(fib->assignedObject))->internalID );
   }
   else{

      return( VOIDENTRY );
   }
}



static int fiberOnObject( xItem *theList, aStar *stars[], fibDB fibers[], double tolerance )

{
   point translatedCoord;
   fibDB *fib;


   if( ( theList->object == VOIDENTRY ) || ( theList->fiber == VOIDENTRY ) ){

      /* we know enough to tell us the rest is pointless! */

      return ( FALSE );
   }

   fib = &fibers[theList->fiber];

   /* apply translations if required */

   buttonToSky( fib, &translatedCoord );

   /* see if the two match up */

   if( distance( &translatedCoord, &(stars[theList->object]->specEncoderCoord) ) < tolerance ){

      return( TRUE );
   }
   else{
      return( FALSE );
   }
}



int userToInternalID( aStar *stars[], int targetID, int *internalID )

{
   int foundFlag = NONE_FOUND;
   int i;

   *internalID = -1;	/* in case we don't find it */

   if(  targetID < 0 )
      return( NONE_FOUND );

   for( i = 0; ( (i< MAXSTARS) && (stars[i] != NULL)); i++ ){

      if( stars[i]->IDnumber == targetID ){

	 if( foundFlag == NONE_FOUND ){

	    *internalID = i;
	    foundFlag = ONE_FOUND;
	 }
	 else{   /* this is a duplicate with this ID */
	    foundFlag = MULTI_FOUND;
	 }
      }
   }
   return( foundFlag );
}



int fiberIsAssignedFop( int fiber, aStar *stars[], fibDB fibers[] )

{
   xItem xList;

   xRefFiber( fiber, &xList, stars, fibers );

   /* fprintf( stderr, "%d %d %d\n", xList.fiber, xList.object, xList.slit ); */

   if( ( fiberIsFop( fiber ) ) && ( xList.object >= 0 ) )
      return( TRUE );
   else
      return( FALSE );
}



int fiberIsFop( int fiber )

{
   if( ( fiber < 0 ) || ( fiber > MAXFIBER ) )
      return( FALSE );

   if( fiberInCable( fiber, getCableIdentifier( "fops" ) ) )
      return( TRUE );
   else
      return( FALSE );
}



int fiberIsSky( int fiber, aStar *stars[], fibDB fibers[] )

{
   return( fiberIsRandomSky( fiber, stars, fibers ) ||
           fiberIsAssignedSky( fiber, stars, fibers ) );
}



int fiberIsAssignedSky( int fiber, aStar *stars[], fibDB fibers[] )


{
   xItem xList;

   xRefFiber( fiber, &xList, stars, fibers );

   if( ( xList.object < 0 ) || ( stars[xList.object] == NULL ) )
      return( FALSE );

   return( (stars[xList.object]->objectType == 'S' ) ||
           (stars[xList.object]->objectType == 's' ) );
}



int fiberIsRandomSky( int fiber, aStar *stars[], fibDB fibers[] )

{
   xItem xList;

   xRefFiber( fiber, &xList, stars, fibers );

   if( xList.object >= 0 )
      return( FALSE );
      

   /* if we get here, we didn't find it explicitly referenced
      --so if it is extended, it's a sky fiber. */
  
   if( fiberParked( fiber ) ){
      return( FALSE );
   }
   else{
      return( TRUE );
   }
}



/*
		fibToFop

   Returns the "slit" position of of a fop, or 
   -1 if the fiber isn't a fop.  Note that
   slit positions are zero indexed for fops,
   and align with the identifiers 'A' to 'Z'
*/

int fibToFop( fibDB *fib )

{

   if( fiberIsFop( fib->fiber_id ) ){
      return( fib->slitPosition );
   }
   else{
      return( -1 );
   }
}




/*
		slitToProbeID

   Convert a "special" slit ID number to an Identifier
   Character for a Field Orientation Probe
*/

char slitToProbeID( int id )

{
   char idChar = 'Z';

   if( id >= 0 ){
      
      idChar = (char)( (int)('A') + id );
   }

   return (idChar);
}



int fiberInCable( int fiber, char cableID )

{
   return( FiBase[fiber].cable == cableID );
}



int fiberBroken( int id )


{
   return( FiBase[ id ].broken == TRUE );
}


int fiberParked( int id )


{
   return( FiBase[ id ].parkedStatus == TRUE );
}
